@@ -33,6 +33,7 @@ Portions Copyright 2012 SN Systems Ltd. All rights reserved.
 #include <stdlib.h> /* exit() free() getenv() realloc() strtoul() */
 #include <string.h> /* memset() strcmp() strlen()
     strncmp() strrchr()*/
+#include <stdio.h> /* FILE decl for dd_esb.h */
 
 /* Windows specific header files */
 #ifdef _WIN32
@@ -48,6 +49,9 @@ Portions Copyright 2012 SN Systems Ltd. All rights reserved.
 #include "dwarf.h"
 #include "libdwarf.h"
 #include "libdwarf_private.h"
+#include "dd_defined_types.h"
+#include "dd_checkutil.h"
+#include "dd_glflags.h"
 #include "dd_globals.h"
 #include "dd_dwconf.h"
 #include "dd_canonical_append.h"
@@ -170,6 +174,62 @@ static int parse_abi(FILE * stream, const char *fname,
     unsigned nest_level);
 static char *get_token(char *cp, struct token_s *outtok);
 
+/*  We insist that the dwarfdump.conf file
+    contain ASCII only and reject if non-ascii
+    characters are found. */
+static int
+is_this_text_file(FILE *stream)
+{
+    unsigned char buf[512];
+    size_t bytesread = 0;
+    unsigned long i = 0;
+    unsigned long curlinelen = 0;
+    unsigned long int maxlinelen = 0;
+
+    bytesread = fread(buf,1,sizeof(buf),stream);
+    if (bytesread != sizeof(buf)) {
+        if (bytesread < 100) {
+            printf("ERROR: Found the configure file is too small to "
+                "be reasonable. Add a few comment lines "
+                "to enlarge it from %lu bytes to 100\n",
+                (unsigned long)bytesread);
+            glflags.gf_count_major_errors++;
+            return FALSE;
+        }
+    }
+    for ( i = 0; i < bytesread; ++i,++curlinelen) {
+        unsigned char c = buf[i];
+
+        if (c == '\n') {
+            if (curlinelen > maxlinelen) {
+                maxlinelen = curlinelen;
+            }
+            curlinelen = 0;
+            continue;
+        }
+        if (c == '\r' || c == '\t') {
+            continue;
+        }
+        if (c >= ' ' && c <= '}') {
+            continue;
+        } else {
+            printf("Found non-ascii character 0x%02x "
+                "and only ascii allowed in dwarfdump.conf\n",
+                c);
+            return FALSE;
+        }
+    }
+    if (maxlinelen > 100) {
+        printf("ERROR: Found maximum configure file"
+            " line length of %lu"
+            " which seems too long to be reasonable\n",
+            maxlinelen);
+        glflags.gf_count_major_errors++;
+        return FALSE;
+    }
+    return TRUE;
+}
+
 /*  This finds a dwarfdump.conf file and
     then parses it.  It updates
     conf_out as appropriate.
@@ -214,19 +274,21 @@ find_conf_file_and_read_config_inner(const char *named_file,
     if (!conf_stream) {
         ++errcount;
         if (!named_file || !strlen(named_file)) {
-            printf("dwarfdump found no dwarfdump.conf file "
+            printf("ERROR: dwarfdump found no dwarfdump.conf file "
                 "in any of the standard places\n"
                 "(precede all arguments "
                 "with option --show-dwarfdump-conf to see "
                 "what file paths tried)\n"
                 "If no print/check arguments also provided "
                 "dwarfdump may silently just stop.\n");
+            glflags.gf_count_major_errors++;
         } else {
-            printf("dwarfdump found no dwarfdump.conf "
+            printf("ERROR: dwarfdump found no dwarfdump.conf "
                 "file \"%s\"\n"
                 "If no print/check arguments also provided "
                 "dwarfdump may silently just stop.\n",
                 sanitized(named_file));
+            glflags.gf_count_major_errors++;
         }
         return FOUND_ERROR;
     }
@@ -236,6 +298,24 @@ find_conf_file_and_read_config_inner(const char *named_file,
     }
     conf_internal->conf_name_used = name_used;
 
+    res = is_this_text_file(conf_stream);
+    if (res == FALSE) {
+        printf("ERROR: dwarfdump configuration "
+            "file \"%s\" is not plain text. Error.\n",
+            sanitized(name_used));
+        glflags.gf_count_major_errors++;
+        fclose(conf_stream);
+        return FOUND_ERROR;
+    }
+    res = fseek(conf_stream,0,SEEK_SET);
+    if (res != 0) {
+        printf("ERROR: dwarfdump configuration "
+            "file \"%s\" fseek to 0 failed. Error\n",
+            sanitized(name_used));
+        glflags.gf_count_major_errors++;
+        fclose(conf_stream);
+        return FOUND_ERROR;
+    }
     /*  And option: lines must come before any abi data. */
     res = find_abi_start(conf_stream, named_abi, &offset, &lineno);
     if (res == FOUND_DONE ||res == FOUND_OPTION) {
@@ -246,10 +326,11 @@ find_conf_file_and_read_config_inner(const char *named_file,
     if (res == FOUND_ERROR) {
         ++errcount;
         fclose(conf_stream);
-        printf("dwarfdump found no usable abi \"%s\" "
+        printf("ERROR: dwarfdump found no usable abi \"%s\" "
             "in file \"%s\".\n",
             named_abi?named_abi:"<not looking for abi>",
             name_used);
+        glflags.gf_count_major_errors++;
         return FOUND_ERROR;
     }
     /* res == FOUND_ABI_START */
@@ -257,8 +338,9 @@ find_conf_file_and_read_config_inner(const char *named_file,
     if (res != 0) {
         ++errcount;
         fclose(conf_stream);
-        printf("dwarfdump seek to %ld offset in %s failed!\n",
+        printf("ERROR: dwarfdump seek to %ld offset in %s failed!\n",
             offset, name_used);
+        glflags.gf_count_major_errors++;
         return FOUND_ERROR;
     }
     parse_abi(conf_stream, name_used, named_abi,
@@ -285,6 +367,7 @@ find_conf_file_and_read_config(const char *named_file,
     res =  find_conf_file_and_read_config_inner(named_file,
         named_abi,
         &conf_internal,0);
+    /*  Nothing in conf_internal need be freed */
     return res;
 }
 
@@ -317,7 +400,7 @@ find_a_file(const char *named_file,
             " configuration as \"%s\"\n", lname);
         fin = fopen(lname, type);
         if (fin) {
-            *name_used = lname;
+            *name_used = makename(lname);
             return fin;
         }
         return 0;
@@ -345,17 +428,16 @@ find_a_file(const char *named_file,
                 "dwarfdump.conf",14);
             lname = szPath;
 
+            if (!lname || !strlen(lname)) {
+                lname="<Impossible file name string>";
+            }
             if (glflags.gf_show_dwarfdump_conf) {
-                if (!lname || !strlen(lname)) {
-                    lname="<Impossible name  string>";
-                }
                 printf("dwarfdump looking for"
                     " configuration as: \"%s\"\n", lname);
             }
-
             fin = fopen(lname, type);
             if (fin) {
-                *name_used = lname;
+                *name_used = makename(lname);
                 return fin;
             }
             /* No file was found */
@@ -368,10 +450,12 @@ find_a_file(const char *named_file,
 
     /* No name given, find a default, if we can. */
     for (i = 0; defaults[i]; ++i) {
+        char buf[2000];
+
+        buf[0] = 0;
         lname = defaults[i];
         if (strncmp(lname, "HOME/", 5) == 0) {
             /* arbitrary size */
-            char buf[2000];
             char *homedir = getenv("HOME");
             if (homedir) {
                 char *cp = _dwarf_canonical_append(buf,
@@ -380,7 +464,7 @@ find_a_file(const char *named_file,
                     /* OOps, ignore this one. */
                     continue;
                 }
-                lname = makename(buf);
+                lname = buf;
             }
         }
         if (glflags.gf_show_dwarfdump_conf) {
@@ -392,7 +476,7 @@ find_a_file(const char *named_file,
         }
         fin = fopen(lname, type);
         if (fin) {
-            *name_used = lname;
+            *name_used = makename(lname);
             return fin;
         }
     }
@@ -543,9 +627,19 @@ get_token(char *cp, struct token_s *outtok)
 {
     char *lcp = skipwhite(cp);
     unsigned tlen = find_token_len(lcp);
+    static int outofmem = FALSE;
 
     outtok->tk_len = tlen;
     if (tlen > 0) {
+        char *src = build_string(tlen, lcp);
+        if (!src) {
+            if (!outofmem) {
+                printf("Dwarfdump out of memory reading "
+                    "dwarfdump.conf and will likely not work.\n");
+            }
+            outofmem = TRUE;
+            return "";
+        }
         outtok->tk_data = build_string(tlen, lcp);
     } else {
         outtok->tk_data = "";
@@ -672,7 +766,7 @@ add_to_reg_table(struct dwconf_s *conf,
         if (!newregs) {
             printf("dwarfdump: unable to malloc table %lu bytes. "
                 " %s line %lu\n", newtabsize, fname, lineno);
-            exit(1);
+            exit(EXIT_FAILURE);
         }
         /* Zero out the new entries. */
         memset((char *) newregs + (oldtabsize), 0,
@@ -1085,12 +1179,13 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
     static int first_time_done = 0;
     struct comtable_s *comtabp = 0;
     int inourabi = FALSE;
+    int unknowntextcount = 0;
 
     if (nest_level > MAX_NEST_LEVEL) {
         ++errcount;
         printf("dwarfdump.conf: includeabi nest "
             "too deep in %s at line %lu\n",
-            fname, lineno);
+            sanitized(fname), lineno);
         return FALSE;
     }
 
@@ -1108,7 +1203,7 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
             ++errcount;
             printf("dwarfdump: end of file or error"
                 " before endabi: in %s, line %lu\n",
-                fname, lineno);
+                sanitized(fname), lineno);
             return FALSE;
         }
         ++lineno;
@@ -1117,13 +1212,21 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
         switch (comtype) {
         case LT_OPTION:
             if (!inourabi) break;
-            parseoption(line, fname, lineno,comtabp);
+            parseoption(line, sanitized(fname), lineno,comtabp);
             break;
         case LT_ERROR:
+            ++unknowntextcount;
             ++errcount;
-            printf ("dwarfdump: Unknown text in %s is \"%s\" "
+            printf("dwarfdump: Unknown text in %s ",
+                sanitized(fname));
+            printf("is \"%s\" "
                 "at line %lu\n",
-                fname, line, lineno);
+                sanitized(line), lineno);
+            if (unknowntextcount > 2) {
+                printf("Too much unknown text. Giving up "
+                    "on the dwarfdump.conf file\n");
+                return FALSE;
+            }
             break;
         case LT_COMMENT:
             break;
@@ -1135,7 +1238,7 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
                 printf("dwarfdump: Encountered beginabi: "
                     "when not expected. "
                     "%s line %lu previous beginabi line %lu\n",
-                    fname,
+                    sanitized(fname),
                     lineno, conf_internal->beginabi_lineno);
             }
             conf_internal->beginabi_lineno = lineno;
@@ -1156,7 +1259,7 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
                     "frame_interface: "
                     "%s line %lu previous frame_interface: "
                     "line %lu\n",
-                    fname, lineno,
+                    sanitized(fname), lineno,
                     conf_internal->frame_interface_lineno);
             }
             conf_internal->frame_interface_lineno = lineno;
@@ -1168,7 +1271,8 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
             if (conf_internal->cfa_reg_lineno > 0) {
                 printf("dwarfdump: Encountered duplicate cfa_reg: "
                     "%s line %lu previous cfa_reg line %lu\n",
-                    fname, lineno, conf_internal->cfa_reg_lineno);
+                    sanitized(fname),
+                    lineno, conf_internal->cfa_reg_lineno);
                 ++errcount;
             }
             conf_internal->cfa_reg_lineno = lineno;
@@ -1182,7 +1286,7 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
                     "initial_reg_value: "
                     "%s line %lu previous initial_reg_value: "
                     "line %lu\n",
-                    fname, lineno,
+                    sanitized(fname), lineno,
                     conf_internal->initial_reg_value_lineno);
                 ++errcount;
             }
@@ -1199,7 +1303,7 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
                     "same_val_reg: "
                     "%s line %lu previous initial_reg_value: "
                     "line %lu\n",
-                    fname, lineno,
+                    sanitized(fname), lineno,
                     conf_internal->initial_reg_value_lineno);
             }
             conf_internal->same_val_reg_lineno = lineno;
@@ -1215,7 +1319,7 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
                     "undefined_val_reg: "
                     "%s line %lu previous initial_reg_value: "
                     "line %lu\n",
-                    fname, lineno,
+                    sanitized(fname), lineno,
                     conf_internal->initial_reg_value_lineno);
             }
             conf_internal->undefined_val_reg_lineno = lineno;
@@ -1227,7 +1331,7 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
             if (conf_internal->reg_table_size_lineno > 0) {
                 printf("dwarfdump: duplicate reg_table_size: "
                     "%s line %lu previous reg_table_size: line %lu\n",
-                    fname, lineno,
+                    sanitized(fname), lineno,
                     conf_internal->reg_table_size_lineno);
                 ++errcount;
             }
@@ -1246,7 +1350,7 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
                     abiname, (unsigned long) conf_internal->regcount,
                     name_reg_table_size,
                     (unsigned long) localconf->cf_table_entry_count,
-                    fname, (unsigned long) lineno);
+                    sanitized(fname), (unsigned long) lineno);
                 ++errcount;
             }
             inourabi = FALSE;
@@ -1292,13 +1396,13 @@ parse_abi(FILE * stream, const char *fname, const char *abiname,
         default:
             printf("dwarfdump internal error,"
                 " impossible line type %d  %s %lu \n",
-                (int) comtype, fname, lineno);
-            exit(1);
+                (int) comtype, sanitized(fname), lineno);
+            exit(EXIT_FAILURE);
         }
     }
     ++errcount;
     printf("End of file, no endabi: found. %s, line %lu\n",
-        fname, lineno);
+        sanitized(fname), lineno);
     return FALSE;
 }
 
