@@ -1,6 +1,6 @@
 /*
 Copyright (C) 2000-2005 Silicon Graphics, Inc. All Rights Reserved.
-Portions Copyright (C) 2008-2020 David Anderson.  All Rights Reserved.
+Portions Copyright (C) 2008-2023 David Anderson.  All Rights Reserved.
 
   This program is free software; you can redistribute it
   and/or modify it under the terms of version 2.1 of the
@@ -662,6 +662,39 @@ static const char _dwarf_errmsgs[DW_DLE_LAST+1][DW_MAX_MSG_LEN] = {
 {"DW_DLE_SYMBOL_SECTION_SIZE_ERROR(490) Corrupt Elf. "
     "Symbols section size bad"},
 {"DW_DLE_PE_SECTION_SIZE_ERROR(491) Corrupt PE object. "
-    "Section size too large."}
+    "Section size too large."},
+{"DW_DLE_DEBUG_ADDR_ERROR(492) Problem reading .debug_addr"},
+{"DW_DLE_NO_SECT_STRINGS(493) Section strings section "
+    "number from header "
+    "is incorrect. Unusable object"},
+{"DW_DLE_TOO_FEW_SECTIONS(494) Sections incomplete, corrupted. "
+    "Unusable object"},
+{"DW_DLE_BUILD_ID_DESCRIPTION_SIZE(495) .note.gnu.build-id section"
+    " corrupt.  Unusable object"},
+{"DW_DLE_BAD_SECTION_FLAGS(496) Some section flags are incorrect."
+    " Unusable object"},
+{"DW_DLE_IMPROPER_SECTION_ZERO(497) Section zero header contents "
+    "incorrect. See Elf ABI.  Unsafe object."},
+{"DW_DLE_INVALID_NULL_ARGUMENT(498) Argument must be a valid "
+    "pointer and non-null"},
+{"DW_DLE_LINE_INDEX_WRONG(499) An index into a line table is "
+    "not valid. Corrupt data."},
+{"DW_DLE_LINE_COUNT_WRONG(500) A count in a line table is "
+    "not valid. Corrupt data."},
+{"DW_DLE_ARITHMETIC_OVERFLOW(501) Arithmetic overflow. "
+    " Corrupt Dwarf." },
+{"DW_DLE_UNIVERSAL_BINARY_ERROR(502) Error reading Mach-O "
+    "uninversal binary head. Corrupt Mach-O object." },
+{"DW_DLE_UNIV_BIN_OFFSET_SIZE_ERROR(503) Offset/size from "
+    "a Mach-O universal binary has an impossible value"},
+{"DW_DLE_PE_SECTION_SIZE_HEURISTIC_FAIL(504) Section size fails "
+    "a heuristic sanity check"},
+{"DW_DLE_LLE_ERROR(505) Generic .debug_loclists read error"},
+{"DW_DLE_RLE_ERROR(506) Generic .debug_rnglists read error"},
+{"DW_DLE_MACHO_SEGMENT_COUNT_HEURISTIC_FAIL(507) "
+    "MachO object seems corrupt"},
+{"DW_DLE_DUPLICATE_NOTE_GNU_BUILD_ID(508) Duplicated section "},
+{"DW_DLE_SYSCONF_VALUE_UNUSABLE(509) sysconf() return is < 200 "
+    "or greater than 100million"}
 };
 #endif /* DWARF_ERRMSG_LIST_H */
